import pytest
from pydantic.error_wrappers import ValidationError

from facturapi.types import BaseQuery


def test_base_query():
    q = BaseQuery(q='Frida')
    assert q.dict() == dict(q='Frida')


def test_base_query_page_size():
    with pytest.raises(ValidationError):
        _ = BaseQuery(q='Frida', limit=51)


def test_base_query_page_min():
    with pytest.raises(ValidationError):
        _ = BaseQuery(q='Frida', page=0)



from enum import Enum


class FileType(str, Enum):
    """Enum to define a file type."""

    pdf = 'pdf'
    xml = 'xml'
    zip = 'zip'


class InvoiceRelation(str, Enum):
    """Relation key in SAT catalogue for related invoices."""

    nota_de_credito = '01'
    nota_de_debito = '02'
    devolucion_de_mercancia = '03'
    sustitucion_de_cfdi_previos = '04'
    traslados_de_mercancia_facturados_previamente = '05'
    factura_por_traslados_previos = '06'
    aplicacion_de_anticipo = '07'
    pagos_en_parcialidades = '08'
    pagos_diferidos = '09'


class InvoiceType(str, Enum):
    """Type of an invoice."""

    ingreso = 'I'
    egreso = 'E'
    traslado = 'T'
    nomina = 'N'
    pago = 'P'


class InvoiceUse(str, Enum):
    """CFDI use code from SAT catalogue."""

    adquisicion_mercancias = 'G01'
    devoluciones_descuentos_bonificaciones = 'G02'
    gastos_en_general = 'G03'
    construcciones = 'I01'
    mobiliario_y_equipo_de_oficina = 'I02'
    equipo_de_transporte = 'I03'
    equipo_de_computo = 'I04'
    dados_troqueles_herramental = 'I05'
    comunicaciones_telefonicas = 'I06'
    comunicaciones_satelitales = 'I07'
    otra_maquinaria = 'I08'
    honorarios_medicos = 'D01'
    gastos_medicos_por_incapacidad = 'D02'
    gastos_funerales = 'D03'
    donativos = 'D04'
    intereses_por_creditos_hipotecarios = 'D05'
    aportaciones_voluntarias_sar = 'D06'
    prima_seguros_gastos_medicos = 'D07'
    gastos_transportacion_escolar = 'D08'
    cuentas_ahorro_pensiones = 'D09'
    servicios_educativos = 'D10'
    por_definir = 'P01'


class PaymentForm(str, Enum):
    """Payment from code according to SAT."""

    efectivo = '01'
    cheque_nominativo = '02'
    transferencia_electronica_de_fondos = '03'
    tarjeta_de_credito = '04'
    monedero_electronico = '05'
    dinero_electronico = '06'
    vales_de_despensa = '06'
    dacion_en_pago = '12'
    pago_por_subrogacion = '13'
    pago_por_consignacion = '14'
    condonacion = '15'
    compensacion = '17'
    novacion = '23'
    confusion = '24'
    remision_de_deuda = '25'
    prescripcion_o_caducidad = '26'
    a_satisfaccion_del_acreedor = '27'
    tarjeta_de_debito = '28'
    tarjeta_de_servicios = '29'
    aplicacion_de_anticipos = '30'
    intermediario_pagos = '31'
    por_definir = '99'


class PaymentMethod(str, Enum):
    """Payment method code according to SAT."""

    contado = 'PUE'
    parcialidades = 'PPD'


class TaxSystemType(str, Enum):
    """Regimen fiscal from SAT catalogue."""

    GRL_PM = "601"  # General de Ley Personas Morales
    PM_FN = "603"  # Personas Morales con Fines no Lucrativos
    SAL_IS = "605"  # Sueldos y Salarios e Ingresos Asimilados a Salarios
    ARR = "606"  # Arrendamiento
    DEM_ING = "608"  # Demás ingresos
    CONS = "609"  # Consolidación
    RES_EXT = "610"  # Residentes en el Extranjero sin Establecimiento Permanente en México # noqa: E501
    DIV = "611"  # Ingresos por Dividendos (socios y accionistas)
    PF_AEP = "612"  # Personas Físicas con Actividades Empresariales y Profesionales # noqa: E501
    INT = "614"  # Ingresos por intereses
    SIN_OBL = "616"  # Sin obligaciones fiscales
    SCP_DIF = "620"  # Sociedades Cooperativas de Producción que optan por diferir sus ingresos # noqa: E501
    INC_FISC = "621"  # Incorporación Fiscal
    AGRO_GAN_SILV_PES = "622"  # Actividades Agrícolas, Ganaderas, Silvícolas y Pesqueras # noqa: E501

    OP_GRUPO_SOC = "623"  # Opcional para Grupos de Sociedades
    COOR = "624"  # Coordinados
    HID = "628"  # Hidrocarburos
    ENAJ_ADQ = "607"  # Régimen de Enajenación o Adquisición de Bienes
    PREF_EM_MULT = "629"  # De los Regímenes Fiscales Preferentes y de las Empresas Multinacionales # noqa: E501
    ENAJ_ACC_BOL = "630"  # Enajenación de acciones en bolsa de valores
    ING_PREM = "615"  # Régimen de los ingresos por obtención de premios
    AE_PLAT_TEC = "625"  # Régimen de las Actividades Empresariales con ingresos a través de Plataformas Tecnológicas # noqa: E501
    RS_CONF = "626"  # Régimen Simplificado de Confianza



__all__ = [
    'Customer',
    'Invoice',
]

from .customers import Customer
from .invoices import Invoice
from .resources import RESOURCES

resource_classes = [
    Customer,
    Invoice,
]
for resource_cls in resource_classes:
    RESOURCES[resource_cls._resource] = resource_cls  # type: ignore



"""
Invoice resource, it includes the class Resource, a request
class to create the resource and a class to represent an
Invoice Item.

"""

import datetime as dt
from typing import ClassVar, Dict, List, Optional, Union, cast

from pydantic import BaseModel
from pydantic.dataclasses import dataclass

from ..types import InvoiceRelation, InvoiceUse, PaymentForm, PaymentMethod
from ..types.general import (
    CustomerBasicInfo,
    ItemPart,
    Namespace,
    ProductBasicInfo,
)
from ..types.queries import InvoiceQuery
from .base import Creatable, Deletable, Downloadable, Queryable, Retrievable
from .customers import Customer, CustomerRequest
from .resources import retrieve_property


class InvoiceItem(BaseModel):
    """
    Class representing an Item from an Invoice.

    Attributes:
        quantity (str): Number of items of this type. Defaults
            to `1`.
        discount (float): Discount on the item price if any.
            Defaults to `0`.
        product (Union[str, ProductBasicInfo, Dict]): Product
            ID, info or request to create a resource.
        custom_keys (List[str]): List of custom product keys.
            Optional.
        complement (str): XML code with additional info to add
            to the invoice. Optional.
        parts (List[ItemParts]): If the concept includes parts.
            Optional.
        property_tax_account (str): 'Predial' number. Optional.

    """

    quantity: Optional[int] = 1
    discount: Optional[float] = 0
    product: Union[
        str, ProductBasicInfo, Dict
    ]  # TO DO: Change Dict for ProductRequest
    customs_keys: Optional[List[str]]
    complement: Optional[str]
    parts: Optional[List[ItemPart]]
    property_tax_account: Optional[str]


class InvoiceRequest(BaseModel):
    """
    This request must be filled to `create` an Invoice.
    It contains all information necessary to create this resource.

    Attributes:
        customer (Union[str, CustomerRequest]): Customer ID or a
            CustomerRequest to create a new one.
        items (List[InvoiceItem]): List of items of the invoice.
        payment_form (PaymentForm): Form of payment.
        payment_method (PaymentMethod): Method of payment. Defaults
            to `PaymentMethod.contado`.
        use (InvoiceUse): Invoice SAT CFDI use. Defaults to
            `InvoiceUse.adquisicion_mercancias`.
        folio_number (int): Internal folio number. Optional.
        series (str): Internal series string. Optional.
        currency (str): Currency of the invoice in ISO format.
            Defaults to `MXN`.
        exchange (float): If a currency is present, the exchange
            value to Mexican Pesos. Defaults to `1.0`.
        conditions (str): Payment conditions. Optional.
        foreign_trade (Dict): Info to add a 'Complemento de Comercio
            Exterior'. Optional.
        related (List[str]): UUID list of related invoices. Optional.
        relation (InvoiceRelation): If related invoices are given,
            their relation key from the SAT catalogue. Optional.
        pdf_custom_section (str): HTML string code to include content
            to the invoice's PDF. Optional
        addenda (str): XML code with Addenda. Optional.
        namespaces (List[Namespace]): If `addenda` or an item complement
            is given, the special namespaces of the XML code. Optional.

    """

    customer: Union[str, CustomerRequest]
    items: List[InvoiceItem]
    payment_form: PaymentForm
    payment_method: Optional[PaymentMethod] = PaymentMethod.contado
    use: Optional[InvoiceUse] = InvoiceUse.adquisicion_mercancias
    folio_number: Optional[int]
    series: Optional[str]
    currency: Optional[str] = 'MXN'
    exchange: Optional[float] = 1.0
    conditions: Optional[str]
    foreign_trade: Optional[Dict]
    related: Optional[List[str]]
    relation: Optional[InvoiceRelation]
    pdf_custom_section: Optional[str]
    addenda: Optional[str]
    namespaces: Optional[Namespace]


@dataclass
class Invoice(Creatable, Deletable, Downloadable, Queryable, Retrievable):
    """Invoice resource

    Resource for an Invoice. It inherits from `Creatable`, `Deletable`,
    `Downloadable`, `Queryable` and `Retrievable`.

    Attributes:
        created_at (datetime.datetime): The datetime in which the
            resource was created.
        livemode (bool): If the resource was created in test or live
            mode.
        status (str): Status of the invoice.
        customer_info (CustomerBasicInfo): Basic info of the Customer.
        customer_uri (str): URI representing how to fetch a Customer
            resource related to the Invoice.
        total (float): Invoice total.
        uuid (str): 'Folio fiscal' assigned by SAT.
        payment_form (PaymentForm): Form of payment of the Invoice.
        items (List[InvoiceItem]): List of items of the Invoice.
        currency (str): Currency of the invoice in ISO format.
        exchange (float): Exchange value to Mexican Pesos.
        cancellation_status (str): If the Invoice was cancelled, the
            status of the cancellation. Optional.
        folio_number (int): Folio number. Optional.
        series (str): Custom series string. Optional. Defaults to `None`.
        related (List[str]): UUID of related invoices. Defaults to
            `None`.
        relation (InvoiceRelation): Relation key from the SAT catalogue.
            Defaults to `None`.

    """

    _resource: ClassVar = 'invoices'
    _relations: ClassVar = ['customer']
    _query_params = InvoiceQuery

    created_at: dt.datetime
    livemode: bool
    status: str
    customer_info: CustomerBasicInfo
    customer_uri: str
    total: float
    uuid: str
    payment_form: PaymentForm
    items: List[InvoiceItem]
    currency: str
    exchange: float
    cancellation_status: Optional[str]
    folio_number: Optional[int]
    series: Optional[str] = None
    related: Optional[List[str]] = None
    relation: Optional[InvoiceRelation] = None

    @classmethod
    def create(cls, data: InvoiceRequest) -> 'Invoice':
        """Create an invoice.

        Args:
            data: All the request data to create an invoice.

        Returns:
            Invoice: The created resource.

        """
        cleaned_data = data.dict(exclude_unset=True, exclude_none=True)
        return cast('Invoice', cls._create(**cleaned_data))

    @classmethod
    def cancel(cls, invoice_id: str, motive: str) -> 'Invoice':
        """Cancel an invoice.

        Calls a DELETE request on invoice resource.

        Args:
            invoice_id: The ID of the invoice to cancel.

        Returns:
            Invoice: The cancelled invoice resource.

        """
        return cast('Invoice', cls._delete(invoice_id, **dict(motive=motive)))

    @property
    def customer(self) -> Customer:
        """Fetch and access Customer resource.

        This property fetches and maps the customer
        related to an invoice so it can be accessed
        through a simple property instead of making a
        manual retrieve.

        Returns:
            Customer: Customer related to the invoice.

        """
        return cast(Customer, retrieve_property(self.customer_uri))



import pytest

import facturapi
from facturapi.resources.customers import CustomerRequest
from facturapi.resources.invoices import InvoiceRequest
from facturapi.types import FileType, PaymentForm
from facturapi.types.exc import MultipleResultsFound, NoResultFound
from facturapi.types.general import CustomerAddress, ItemPart


@pytest.mark.vcr
def test_create_invoice():
    invoice_request = InvoiceRequest(
        customer=CustomerRequest(
            legal_name='Remedios Varo',
            tax_id='VAUR631216M55',
            email='remedios@varo.com',
            tax_system='625',
            address=CustomerAddress(
                street='Colima',
                exterior='196',
                interior='1',
                neighborhood='Roma',
                zip='06700',
                city='Ciudad de México',
                municipality='Cuauhtémoc',
                state='Ciudad de México',
            ),
        ),
        items=[
            dict(
                product=dict(
                    description='Producto Test',
                    product_key='50202201',
                    price=42.05,
                ),
                quantity=2,
                discount=0.1,
                parts=[
                    ItemPart(
                        description='Parte 1',
                        product_key='50202201',
                    )
                ],
            ),
        ],
        payment_form=PaymentForm.tarjeta_de_credito,
    )

    invoice = facturapi.Invoice.create(data=invoice_request)
    assert invoice.id
    assert invoice.created_at
    assert invoice.status
    assert invoice.uuid
    assert invoice.items

    invoice_dict = invoice.to_dict()
    assert isinstance(invoice_dict, dict)
    assert invoice_dict['id']
    assert invoice_dict['created_at']
    assert invoice_dict['uuid']
    assert invoice_dict['items']


@pytest.mark.vcr
def test_retrieve_invoice():
    invoice_request = InvoiceRequest(
        customer='5fc5aa9938e6a2001b31aa21',
        items=[
            dict(
                product=dict(
                    description='Producto Test',
                    product_key='50202201',
                    price=42.05,
                ),
            ),
        ],
        payment_form=PaymentForm.tarjeta_de_credito,
    )
    invoice = facturapi.Invoice.create(data=invoice_request)
    retrieved_invoice = facturapi.Invoice.retrieve(id=invoice.id)

    assert retrieved_invoice.id == invoice.id
    assert retrieved_invoice.created_at == invoice.created_at
    assert retrieved_invoice.status == invoice.status
    assert retrieved_invoice.uuid == invoice.uuid


@pytest.mark.vcr
def test_cancel_invoice():
    invoice_id = '63fe4be4e87ce2001b19077f'
    invoice = facturapi.Invoice.retrieve(id=invoice_id)

    cancelled_invoice = facturapi.Invoice.cancel(
        invoice_id=invoice.id, motive='01'
    )

    assert cancelled_invoice.id == invoice.id
    assert cancelled_invoice.status != invoice.status

    invoice.refresh()
    assert cancelled_invoice.status == invoice.status
    assert cancelled_invoice.cancellation_status == 'accepted'
    assert invoice.cancellation_status == 'accepted'


@pytest.mark.vcr
def test_download_invoice():
    invoice_id = 'INVOICE01'
    invoice_bytes = facturapi.Invoice.download(
        id=invoice_id, file_type=FileType.pdf
    )

    assert invoice_bytes
    assert type(invoice_bytes) == bytes


@pytest.mark.vcr
def test_download_invoice_xml():
    invoice_id = 'INVOICE01'
    invoice_bytes = facturapi.Invoice.download(
        id=invoice_id, file_type=FileType.xml
    )

    assert invoice_bytes
    assert type(invoice_bytes) == bytes


@pytest.mark.vcr
def test_download_invoice_zip():
    invoice_id = 'INVOICE01'
    invoice_bytes = facturapi.Invoice.download(
        id=invoice_id, file_type=FileType.zip
    )

    assert invoice_bytes
    assert type(invoice_bytes) == bytes


@pytest.mark.vcr
def test_invoice_customer_property():
    invoice_id = 'INVOICE01'
    invoice = facturapi.Invoice.retrieve(id=invoice_id)
    assert invoice.id

    customer = invoice.customer
    assert customer.id == invoice.customer_info.id


@pytest.mark.vcr
def test_query_invoice_one():
    invoice = facturapi.Invoice.one()
    assert invoice.id


@pytest.mark.vcr
def test_query_invoice_one_multiple():
    with pytest.raises(MultipleResultsFound):
        _ = facturapi.Invoice.one()


@pytest.mark.vcr
def test_query_invoice_one_no_found():
    with pytest.raises(NoResultFound):
        _ = facturapi.Invoice.one(q='Cordelia Urueta Sierra')


@pytest.mark.vcr
def test_query_invoice_first():
    invoice = facturapi.Invoice.first()
    assert invoice is not None
    assert invoice.id


@pytest.mark.vcr
def test_query_invoice_first_none():
    invoice = facturapi.Invoice.first(q='Cordelia Urueta Sierra')
    assert invoice is None


@pytest.mark.vcr
def test_query_invoice_count():
    count = facturapi.Invoice.count()
    assert count == 1


@pytest.mark.vcr
def test_query_invoice_all():
    all_invoices = facturapi.Invoice.all(limit=1)

    for invoice in all_invoices:
        assert invoice.id



import pytest

import facturapi
from facturapi.resources.customers import (
    CustomerRequest,
    CustomerUpdateRequest,
)
from facturapi.types.exc import MultipleResultsFound, NoResultFound
from facturapi.types.general import CustomerAddress


@pytest.mark.vcr
def test_create_customer():
    customer_request = CustomerRequest(
        legal_name='Cordelia Urueta Sierra',
        tax_id='UUSC9509162G7',
        tax_system='625',
        email='cordelia@urueta.com',
        phone='5511223344',
        address=CustomerAddress(
            street='Colima',
            exterior='196',
            interior='1',
            neighborhood='Roma',
            zip='06700',
            city='Ciudad de México',
            municipality='Cuauhtémoc',
            state='Ciudad de México',
        ),
    )

    customer = facturapi.Customer.create(data=customer_request)
    assert customer.id
    assert customer.created_at
    assert customer.legal_name == 'CORDELIA URUETA SIERRA'
    assert customer.tax_id == 'UUSC9509162G7'
    assert customer.email == 'cordelia@urueta.com'

    customer_dict = customer.to_dict()
    assert isinstance(customer_dict, dict)
    assert customer_dict['id']
    assert customer_dict['created_at']
    assert customer_dict['legal_name']
    assert customer_dict['tax_id']
    assert customer_dict['email']


@pytest.mark.vcr
def test_retrieve_customer():
    customer_request = CustomerRequest(
        legal_name='Leonora Carrington',
        tax_id='CAML9004069U0',
        email='leonora@test.com',
        tax_system='625',
        address=CustomerAddress(
            street='Colima',
            exterior='196',
            interior='1',
            neighborhood='Roma',
            zip='06700',
            city='Ciudad de México',
            municipality='Cuauhtémoc',
            state='Ciudad de México',
        ),
    )
    customer = facturapi.Customer.create(data=customer_request)

    retrieved_customer = facturapi.Customer.retrieve(id=customer.id)

    assert retrieved_customer.id == customer.id
    assert retrieved_customer.created_at == customer.created_at
    assert retrieved_customer.legal_name == customer.legal_name
    assert retrieved_customer.tax_id == customer.tax_id


@pytest.mark.vcr
def test_update_customer():
    customer_id = '63fe41cee87ce2001b18484a'
    customer = facturapi.Customer.retrieve(id=customer_id)
    update_data = CustomerUpdateRequest(email='remedios@pintora.com')

    updated_customer = facturapi.Customer.update(
        id=customer_id, data=update_data
    )

    assert customer.email != updated_customer.email
    assert updated_customer.email == 'remedios@pintora.com'

    # Test refreshing an object
    customer.refresh()
    assert customer.email == updated_customer.email


@pytest.mark.vcr
def test_query_customer_one():
    customer = facturapi.Customer.one(q='Remedios')
    assert customer.id


@pytest.mark.vcr
def test_query_customer_one_multiple():
    with pytest.raises(MultipleResultsFound):
        _ = facturapi.Customer.one()


@pytest.mark.vcr
def test_query_customer_one_no_found():
    with pytest.raises(NoResultFound):
        _ = facturapi.Customer.one(q='Diego Rivera')


@pytest.mark.vcr
def test_query_customer_first():
    customer = facturapi.Customer.first()
    assert customer is not None
    assert customer.id


@pytest.mark.vcr
def test_query_customer_first_none():
    customer = facturapi.Customer.first(q='Diego Rivera')
    assert customer is None


@pytest.mark.vcr
def test_query_customer_count():
    count = facturapi.Customer.count()
    assert count == 2


@pytest.mark.vcr
def test_query_customer_all():
    all_customers = facturapi.Customer.all()

    for customer in all_customers:
        assert customer.id



import pytest

from facturapi.http.client import Client
from facturapi.types.exc import FacturapiResponseException


def test_auth_without_env():
    client = Client()
    assert not client.api_key


def test_auth_configure():
    client = Client()
    client.configure('some_api_key')
    assert client.api_key == 'some_api_key'


@pytest.mark.vcr
def test_invalid_auth():
    client = Client()
    with pytest.raises(FacturapiResponseException) as e:
        client.get('/invoices')
    assert e.value.status_code == 401
    assert str(e.value)


@pytest.mark.usefixtures('facturapi_creds')
def test_env_auth():
    client = Client()
    assert client.api_key == 'some_api_key'


@pytest.mark.usefixtures('facturapi_creds')
def test_override_with_configure():
    client = Client()
    client.configure('some_api_key2')
    assert client.api_key == 'some_api_key2'



import re
from typing import Dict, cast

from .base import Retrievable

ENDPOINT_RE = re.compile(r'(?P<resource>[a-z]+)/(?P<id>.+)$')
RESOURCES: Dict[str, Retrievable] = {}  # set in ./__init__.py after imports


def retrieve_property(uri: str) -> Retrievable:
    m = ENDPOINT_RE.match(uri)
    if not m:
        raise ValueError(f'uri is not a valid format: {uri}')
    resource, id_ = m.groups()
    return cast(Retrievable, RESOURCES[resource].retrieve(id_))



from dataclasses import dataclass
from typing import Any, Dict


class FacturapiException(Exception):
    ...


class NoResultFound(FacturapiException):
    """No results where found"""


class MultipleResultsFound(FacturapiException):
    """One result was expected but multiple were returned"""


@dataclass
class FacturapiResponseException(FacturapiException):
    json: Dict[str, Any]
    status_code: int

    def __str__(self) -> str:
        return repr(self)



__all__ = [
    '__version__',
    'Customer',
    'Invoice',
    'configure',
]

from .http import client
from .resources import Customer, Invoice
from .version import __version__

configure = client.configure



from typing import List, Optional

from pydantic import BaseModel

from .validators import sanitize_dict


class CustomerAddress(BaseModel):
    """Address of a customer.

    Attributes:
        street (str): Street.
        exterior (str): Exterior place number.
        interior (str): Interior place number.
        neighborhood (str): 'Colonia'.
        zip (str): Postal code.
        city (str): City.
        municipality (str): 'Municipio or Alcaldía'.
        state (str): State of the address.
        country (str): Country.

    """

    street: Optional[str]
    exterior: Optional[str]
    interior: Optional[str]
    neighborhood: Optional[str]
    zip: Optional[str]
    city: Optional[str]
    municipality: Optional[str]
    state: Optional[str]
    country: Optional[str]


class CustomerBasicInfo(BaseModel):
    """Customer's basic info"""

    id: str
    legal_name: str
    tax_id: str


class ItemPart(BaseModel):
    """Defines a part of an invoice item."""

    description: str
    product_key: str
    quantity: Optional[int] = 1
    sku: Optional[str]
    unit_price: Optional[float]
    customs_keys: Optional[List[str]]


class Namespace(BaseModel):
    """Namespace for spceial XML namespaces for an invoice."""

    prefix: Optional[str]
    uri: Optional[str]
    schema_location: Optional[str]


class ProductBasicInfo(BaseModel):
    """Product's basic info."""

    id: str
    unit_name: str
    description: str


class SanitizedDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        sanitize_dict(self)



import os
from typing import Any, Dict, MutableMapping, Optional, Union
from urllib.parse import urljoin

import requests
from requests import Response

from ..types.exc import FacturapiResponseException
from ..version import CLIENT_VERSION

API_HOST = 'www.facturapi.io/v2'


class Client:
    """Client to perform http requests to Facturapi.

    By default it inits and uses an `API_KEY` configured as
    an environment variable `FACTURAPI_KEY`, if the key
    is going to be set latter, the method `configure()`
    can be used.

    Attributes:
        host (str): Base URL to perform requests.
        session (requests.Session): The requests session used
            to perform requests.
        api_key (str): API KEY for Facturapi

    """

    host: str = API_HOST
    session: requests.Session

    def __init__(self):
        self.session = requests.Session()
        self.session.headers.update(
            {
                'User-Agent': f'facturapi-python/{CLIENT_VERSION}',
                'Content-Type': 'application/json',
            }
        )

        # Auth
        self.api_key = os.getenv('FACTURAPI_KEY', '')
        self.session.auth = (self.api_key, '')

    def configure(self, api_key: str):
        """Configure the http client.

        Import the client and configure it passing the `API_KEY`
        instead of configuring it as an environment variable.

        Args:
            api_key: Facturapi `API_KEY`

        """
        self.api_key = api_key
        self.session.auth = (self.api_key, '')

    def get(
        self,
        endpoint: str,
        params: Union[None, bytes, MutableMapping[str, str]] = None,
    ) -> Dict[str, Any]:
        """Performs GET request to Facturapi."""
        return self.request('get', endpoint, params=params)

    def post(self, endpoint: str, data: Dict[str, Any]) -> Dict[str, Any]:
        """Performs POST request to Facturapi."""
        return self.request('post', endpoint, data=data)

    def put(self, endpoint: str, data: Dict[str, Any]) -> Dict[str, Any]:
        """Performs PUT request to Facturapi."""
        return self.request('put', endpoint, data=data)

    def delete(self, endpoint: str) -> Dict[str, Any]:
        """Performs DELETE request to Facturapi."""
        return self.request('delete', endpoint)

    def request(
        self,
        method: str,
        endpoint: str,
        params: Union[None, bytes, MutableMapping[str, str]] = None,
        data: Optional[Dict[str, Union[int, str]]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """Performs a request to Facturapi.

        Given a `method` and `endpoint`, perform a request to
        Facturapi.

        Args:
            method: HTTP method of the request.
            endpoint: Endpoint to make the request to.
            params: URL encoded parameters. Defaults to `None`.
            data: JSON data to be sent. Defaults to `None`.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Dict[str, Any]: JSON of the request's response.

        Raises:
            FacturapiResponseException: If response is not
                successful.

        """
        response = self.session.request(
            method=method,
            url=('https://' + self.host + urljoin('/', endpoint)),
            json=data,
            params=params,
            **kwargs,
        )
        self._check_response(response)
        return response.json()

    def download_request(
        self,
        endpoint: str,
        **kwargs,
    ) -> bytes:
        """Performs a GET request handling a file download.

        Args:
            endpoint: Endpoint to make the request to.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            bytes: Bytes of the requested file.

        Raises:
            FacturapiResponseException: If response is not
                successful.

        """
        response = self.session.request(
            method='GET',
            url=('https://' + self.host + urljoin('/', endpoint)),
            **kwargs,
        )
        self._check_response(response)
        return response.content

    @staticmethod
    def _check_response(response: Response):
        if not response.ok:
            raise FacturapiResponseException(
                json=response.json(),
                status_code=response.status_code,
            )



import datetime as dt
from typing import Any, Dict, Optional, Union

from pydantic import BaseModel, Extra
from pydantic.types import ConstrainedInt

MAX_PAGE_SIZE = 50
MIN_PAGE = 1


class PageSize(ConstrainedInt):
    gt = 0
    le = MAX_PAGE_SIZE


class Page(ConstrainedInt):
    gt = MIN_PAGE


class DateFilter(BaseModel):
    """Model for a date filter query.

    Defines possible filters for date values.
    Only a few resources' query can be filtered
    through date. For more details check out the docs.

    Attributes:
        gt: Filter for greater than the given date.
        gte: Filter for greater than or equals to the
            given date.
        lt: Filter for lesser than the given date.
        lte: Filter for lesser than or equals to the
        given date.

    """

    gt: Optional[Union[str, dt.datetime]]
    gte: Optional[Union[str, dt.datetime]]
    lt: Optional[Union[str, dt.datetime]]
    lte: Optional[Union[str, dt.datetime]]


class BaseQuery(BaseModel):
    """Base query to query for resources.

    Defines all the possible arguments to query
    for a given resource.

    Attributes:
        q: Query text, its value may depend on what you are
            querying for. Check out Facturapi's docs for more
            details.
        limit: Max number of results thrown in a query.
            Defaults to 50, its maximum value.
        page: Number of page to query. Defaults to 1.
        date: Object of type `DateFilter` that can contain one
            or more date filters.

    """

    q: Optional[str]
    limit: Optional[PageSize] = PageSize(MAX_PAGE_SIZE)
    page: Optional[Page] = Page(MIN_PAGE)
    date: Optional[DateFilter]

    class Config:
        extra = Extra.forbid

    def dict(self, *args, **kwargs) -> Dict[str, Any]:
        kwargs.setdefault('exclude_none', True)
        kwargs.setdefault('exclude_unset', True)
        d = super().dict(*args, **kwargs)
        return d


class InvoiceQuery(BaseQuery):
    motive: Optional[str]



import datetime as dt
from enum import Enum

from pydantic.dataclasses import dataclass

from facturapi.types.general import SanitizedDict


class TEnum(Enum):
    zero = 0


@dataclass
class TClass:
    ein: int

    def to_dict(self):
        return dict(ein=self.ein, zwei='dos')


def test_sinitized_dict():
    now = dt.datetime.now()
    utc_now = now.astimezone(dt.timezone.utc)

    utc_date = dt.datetime.now(dt.timezone.utc)

    class_with_dict = TClass(ein=1)

    assert SanitizedDict(
        date=now,
        hello='there',
        many_numbers=class_with_dict,
        number=TEnum.zero,
        utc_date=utc_date,
    ) == dict(
        date=utc_now.isoformat(),
        hello='there',
        many_numbers=dict(
            ein=1,
            zwei='dos',
        ),
        number=0,
        utc_date=utc_date.isoformat(),
    )



__all__ = [
    'BaseQuery',
    'DateFilter',
    'FacturapiResponseException',
    'FileType',
    'InvoiceRelation',
    'InvoiceType',
    'InvoiceUse',
    'PaymentForm',
    'PaymentMethod',
    'SanitizedDict',
    'exc',
    'general',
    'validators',
]

from . import exc, general, validators
from .enums import (
    FileType,
    InvoiceRelation,
    InvoiceType,
    InvoiceUse,
    PaymentForm,
    PaymentMethod,
)
from .queries import BaseQuery, DateFilter



"""Base module for resources.

This module represent all the base code for resources,
the resources can inherit behaviour from these classes to
perform requests and actions to the API.
"""

from dataclasses import asdict, fields
from typing import Any, ClassVar, Dict, Generator, List, Optional
from urllib.parse import urlencode

from pydantic.dataclasses import dataclass

from ..http import client
from ..types import BaseQuery, FileType
from ..types.exc import MultipleResultsFound, NoResultFound
from ..types.general import SanitizedDict


@dataclass
class Resource:
    """Generic resource from Facturapi.

    Generic Resource class used by Facturapi resources.

    Attributes:
        _resource (ClassVar[str]): Name of the resource the
            class corresponds to.
        id (str): ID of the resource.

    """

    _resource: ClassVar[str]
    _relations: ClassVar[List[str]] = []

    id: str

    # purely for MyPy
    def __init__(self, **_):  # pragma: no cover
        ...

    @classmethod
    def _from_dict(cls, obj_dict: Dict[str, Any]) -> 'Resource':
        cls._filter_excess_fields(obj_dict)
        return cls(**obj_dict)

    @classmethod
    def _filter_excess_fields(cls, obj_dict: Dict[str, Any]) -> None:
        """
        dataclasses don't allow __init__ to be called with excess fields.
        This method allows the API to add fields in the response body without
        breaking the client.
        Also if data of a relation is found, the data is rearranged so the
        library can map relations easier.

        """
        excess = set(obj_dict.keys()) - {f.name for f in fields(cls)}
        for f in excess:
            if f in cls._relations:
                id_ = obj_dict[f]['id']
                obj_dict[f'{f}_uri'] = f'{f}s/{id_}'
                obj_dict[f'{f}_info'] = obj_dict[f]
            del obj_dict[f]

    def to_dict(self) -> Dict:
        return asdict(self, dict_factory=SanitizedDict)


class Retrievable(Resource):
    """Generic Retrievable class.

    Used by resources that can be retrieved, i.e: resources that can
    be retrieved by a GET request to its ID.

    """

    @classmethod
    def retrieve(cls, id: str) -> Resource:
        """Retrieve a resource given its ID

        Performs a GET request with the ID.

        Args:
            id: The ID of the resource

        Returns:
            Resource: The resource retrieved.

        """
        response = client.get(f'/{cls._resource}/{id}')
        return cls._from_dict(response)

    def refresh(self):
        """Refresh a resource

        Refresh resource's data to be sure its the latest. It
        performs a GET request on the resource.

        Returns:
            Resource: The refreshed resource.

        """
        new = self.retrieve(self.id)
        for attr, value in new.__dict__.items():
            setattr(self, attr, value)


class Downloadable(Resource):
    """Generic Downloadable class.

    Used by resources that can be downloaded as a file.

    """

    @classmethod
    def download(cls, id: str, file_type: FileType) -> bytes:
        """Download a file from resource.

        Performs a GET request to download a file given a
        resource ID.

        Args:
            id: The ID of the resource.
            file_type: Type of the file to be downloaded.
                (zip, pdf or xml).

        Returns:
            bytes: Bytes of the file.

        """
        return client.download_request(
            f'/{cls._resource}/{id}/{file_type.value}'
        )


class Creatable(Resource):
    """Generic Creatable class.

    Used by resources that can be created.

    """

    @classmethod
    def _create(cls, **data) -> Resource:
        """Create a resource

        Performs a POST request with the data.

        Args:
            data (dict): Data from the resource to create.

        Returns:
            Resource: The created resource.

        """
        response = client.post(cls._resource, data)
        return cls._from_dict(response)


class Updatable(Resource):
    """Generic Updatable class.

    Used by resources that can be updated.

    """

    @classmethod
    def _update(cls, id: str, **data) -> Resource:
        """Update an specific resource with new data.

        Performs a PUT request with the updated data.

        Args:
            id: The ID of the resource.

        Returns:
            Resource: The updated resource.

        """
        response = client.put(f'/{cls._resource}/{id}', data)
        return cls._from_dict(response)


class Deletable(Resource):
    """Generic Deletable class.

    Used by resources that can be deleted.

    """

    _query_params: ClassVar = BaseQuery

    @classmethod
    def _delete(cls, id: str, **query_params) -> Resource:
        """Delete an specific resource.

        Performs a DELETE request on the ID.

        Args:
            id: The ID of the resource to delete.
            query_params: Send aditional information over uri

        Returns:
            Resource: The deleted resource.

        """
        q = cls._query_params(**query_params)
        response = client.delete(
            f'/{cls._resource}/{id}?{urlencode(q.dict())}'
        )
        return cls._from_dict(response)


@dataclass
class Queryable(Resource):
    """Generic Queryable class.

    Used by resources that can be queried in lists. Refer to this
    class to see the query actions that can be performed on a
    resource.

    Attributes:
        _query_params: A class with the parameters that
            can be queried.
    """

    _query_params: ClassVar = BaseQuery

    @classmethod
    def one(cls, **query_params) -> Resource:
        """Retrieve only one resource given a query.

        Given a query, retrieve one and only one resource. If more
        than one resource or none are found, this method throws an
        exception.

        Args:
            **query_params (dict): Arbitrary query keyword arguments.

        Raises:
            NoResultFound: If no result is found.
            MultipleResultsFound: If more than one result is found.

        Returns:
            Resource: The one resource queried.

        """
        q = cls._query_params(limit=2, **query_params)
        response = client.get(cls._resource, q.dict())
        items = response['data']
        len_items = len(items)
        if not len_items:
            raise NoResultFound
        if len_items > 1:
            raise MultipleResultsFound
        return cls._from_dict(items[0])

    @classmethod
    def first(cls, **query_params) -> Optional[Resource]:
        """Retrieve the first resource found given a query or none.

        Args:
            **query_params (dict): Arbitrary query keyword arguments.

        Returns:
            Optional[Resource]: The first resource queried or `None`
                if none found.

        """
        q = cls._query_params(limit=1, **query_params)
        response = client.get(cls._resource, q.dict())
        try:
            item = response['data'][0]
        except IndexError:
            rv = None
        else:
            rv = cls._from_dict(item)
        return rv

    @classmethod
    def count(cls, **query_params) -> int:
        """Get the total number of results given a query.

        Args:
            **query_params (dict): Arbitrary query keyword arguments.

        Returns:
            int: The total count of results.

        """
        q = cls._query_params(**query_params)
        response = client.get(cls._resource, q.dict())
        items = response['data']
        return len(items)

    @classmethod
    def all(cls, **query_params) -> Generator[Resource, None, None]:
        """Retrieve all resources given a query.

        All the returned resources are paginated, the method `yields`
        the first page of results and if more are found; then it
        continues to `yield` pages until all results are queried.

        Args:
            **query_params (dict): Arbitrary query keyword arguments.

        Returns:
            Generator: A generator containing the queried results.

        """
        q = cls._query_params(**query_params)
        next_page_uri = f'{cls._resource}?{urlencode(q.dict())}'
        current_page = 1
        while next_page_uri:
            page = client.get(next_page_uri)
            yield from (cls._from_dict(item) for item in page['data'])
            next_page_uri = ''
            if current_page < page['total_pages']:
                current_page += 1
                q.page = current_page
                next_page_uri = f'{cls._resource}?{urlencode(q.dict())}'



"""
Customer resource, it includes the class Resource and two request
classes to create and update the resource.

"""

import datetime as dt
from typing import ClassVar, Optional, cast

from pydantic import BaseModel
from pydantic.dataclasses import dataclass

from ..types.enums import TaxSystemType
from ..types.general import CustomerAddress
from .base import Creatable, Queryable, Retrievable, Updatable


class CustomerRequest(BaseModel):
    """
    This request must be filled to `create` a Customer.
    It contains all information necessary to create this resource.

    Attributes:
        legal_name (str): Full name of the customer.
        tax_id (str): RFC of the customer.
        email (str): Email of the customer.
        phone (str): Phone of the customer. Optional.
        address (CustomerAddress): Address object of the customer. Optional.

    """

    legal_name: str
    tax_id: str
    tax_system: TaxSystemType
    email: str
    phone: Optional[str]
    address: CustomerAddress


class CustomerUpdateRequest(BaseModel):
    """
    This request must be filled to `update` a Customer.
    It contains all information necessary to update this resource.

    Attributes:
        legal_name (str): Full name of the customer. Optional.
        tax_id (str): RFC of the customer. Optional.
        email (str): Email of the customer. Optional.
        phone (str): Phone of the customer. Optional.
        address (CustomerAddress): Address object of the customer. Optional.

    """

    legal_name: Optional[str]
    tax_id: Optional[str]
    tax_system: Optional[TaxSystemType]
    email: Optional[str]
    phone: Optional[str]
    address: Optional[CustomerAddress]


@dataclass
class Customer(Creatable, Queryable, Retrievable, Updatable):
    """Customer resource

    Resource for a Customer. It inherits from `Creatable`, `Queryable`,
    `Retrievable` and `Updatable`.

    Attributes:
        created_at (datetime.datetime): The datetime in which the
            resource was created.
        livemode (bool): If the resource was created in test or live
            mode.
        legal_name (str): Name of the customer.
        tax_id (str): RFC of the customer.
        email (str): Email of the customer.
        address (CustomerAddress): Address data of the model. Optional.
        tax_system (TaxSystemType): Enum from catalogue of SAT
        phone (str): Phone of the customer. Defaults to `None`.
    """

    _resource: ClassVar = 'customers'

    created_at: dt.datetime
    livemode: bool
    legal_name: str
    tax_id: str
    email: str
    address: CustomerAddress
    tax_system: Optional[TaxSystemType] = None
    phone: Optional[str] = None

    @classmethod
    def create(cls, data: CustomerRequest) -> 'Customer':
        """Create a customer.

        Args:
            data: All the request data to create a customer.

        Returns:
            Customer: The created customer resource.

        """
        cleaned_data = data.dict(exclude_unset=True, exclude_none=True)
        return cast('Customer', cls._create(**cleaned_data))

    @classmethod
    def update(cls, id: str, data: CustomerUpdateRequest) -> 'Customer':
        """Update a customer.

        Args:
            id: ID of the customer to be updated.
            data: Data to be updated.

        Returns:
            Customer: The udpated customer resource.

        """
        cleaned_data = data.dict(exclude_unset=True, exclude_none=True)
        return cast('Customer', cls._update(id=id, **cleaned_data))



import datetime as dt
from enum import Enum
from typing import Any


def sanitize_dict(d: dict):
    for k, v in d.items():
        d[k] = sanitize_item(v)


def sanitize_item(item: Any) -> Any:
    """Sanitizes a value to use simple data.

    Args:
        item: Item to be sanatized.
        default: Optional function to be used when there is
            no case for an item type. Defaults to `None`.

    Returns:
        Any: Sanitized value.

    """
    if isinstance(item, dt.date):
        if isinstance(item, dt.datetime) and not item.tzinfo:
            return item.astimezone(dt.timezone.utc).isoformat()
        else:
            return item.isoformat()
    elif isinstance(item, Enum):
        return item.value
    elif hasattr(item, 'to_dict'):
        return item.to_dict()
    else:
        return item
